import { D2, MultiSet, output } from "@electric-sql/d2mini";
import { createCollection } from "../collection.js";
import { compileQueryPipeline } from "./pipeline-compiler.js";
function compileQuery(queryBuilder) {
  return new CompiledQuery(queryBuilder);
}
class CompiledQuery {
  constructor(queryBuilder) {
    this.state = `compiled`;
    this.unsubscribeCallbacks = [];
    const query = queryBuilder._query;
    const collections = query.collections;
    if (!collections) {
      throw new Error(`No collections provided`);
    }
    this.inputCollections = collections;
    const graph = new D2();
    const inputs = Object.fromEntries(
      Object.entries(collections).map(([key]) => [key, graph.newInput()])
    );
    const sync = ({
      begin,
      write,
      commit,
      collection
    }) => {
      compileQueryPipeline(
        query,
        inputs
      ).pipe(
        output((data) => {
          begin();
          data.getInner().reduce((acc, [[key, value], multiplicity]) => {
            const changes = acc.get(key) || {
              deletes: 0,
              inserts: 0,
              value
            };
            if (multiplicity < 0) {
              changes.deletes += Math.abs(multiplicity);
            } else if (multiplicity > 0) {
              changes.inserts += multiplicity;
              changes.value = value;
            }
            acc.set(key, changes);
            return acc;
          }, /* @__PURE__ */ new Map()).forEach((changes, rawKey) => {
            const { deletes, inserts, value } = changes;
            const valueWithKey = { ...value, _key: rawKey };
            if (inserts && deletes === 0) {
              write({
                value: valueWithKey,
                type: `insert`
              });
            } else if (
              // Insert & update(s) (updates are a delete & insert)
              inserts > deletes || // Just update(s) but the item is already in the collection (so
              // was inserted previously).
              inserts === deletes && collection.has(valueWithKey._key)
            ) {
              write({
                value: valueWithKey,
                type: `update`
              });
            } else if (deletes > 0) {
              write({
                value: valueWithKey,
                type: `delete`
              });
            } else {
              throw new Error(
                `This should never happen ${JSON.stringify(changes)}`
              );
            }
          });
          commit();
        })
      );
      graph.finalize();
    };
    this.graph = graph;
    this.inputs = inputs;
    const compare = query.orderBy ? (val1, val2) => {
      const x = val1;
      const y = val2;
      if (x._orderByIndex < y._orderByIndex) {
        return -1;
      } else if (x._orderByIndex > y._orderByIndex) {
        return 1;
      } else {
        return 0;
      }
    } : void 0;
    this.resultCollection = createCollection({
      getKey: (val) => {
        return val._key;
      },
      gcTime: 0,
      startSync: true,
      compare,
      sync: {
        sync
      }
    });
  }
  get results() {
    return this.resultCollection;
  }
  sendChangesToInput(inputKey, changes, getKey) {
    const input = this.inputs[inputKey];
    const multiSetArray = [];
    for (const change of changes) {
      const key = getKey(change.value);
      if (change.type === `insert`) {
        multiSetArray.push([[key, change.value], 1]);
      } else if (change.type === `update`) {
        multiSetArray.push([[key, change.previousValue], -1]);
        multiSetArray.push([[key, change.value], 1]);
      } else {
        multiSetArray.push([[key, change.value], -1]);
      }
    }
    input.sendData(new MultiSet(multiSetArray));
  }
  runGraph() {
    this.graph.run();
  }
  start() {
    if (this.state === `running`) {
      throw new Error(`Query is already running`);
    } else if (this.state === `stopped`) {
      throw new Error(`Query is stopped`);
    }
    Object.entries(this.inputCollections).forEach(([key, collection]) => {
      const unsubscribe = collection.subscribeChanges(
        (changes) => {
          this.sendChangesToInput(key, changes, collection.config.getKey);
          this.runGraph();
        },
        { includeInitialState: true }
      );
      this.unsubscribeCallbacks.push(unsubscribe);
    });
    this.runGraph();
    this.state = `running`;
    return () => {
      this.stop();
    };
  }
  stop() {
    this.unsubscribeCallbacks.forEach((unsubscribe) => unsubscribe());
    this.unsubscribeCallbacks = [];
    this.state = `stopped`;
  }
}
export {
  CompiledQuery,
  compileQuery
};
//# sourceMappingURL=compiled-query.js.map
