import { evaluateOperandOnNamespacedRow } from "./extractors.js";
import { isValueInArray, convertLikeToRegex, compareValues } from "./utils.js";
function evaluateWhereOnNamespacedRow(namespacedRow, where, mainTableAlias, joinedTableAlias) {
  return where.every((item) => {
    if (typeof item === `function`) {
      return item(namespacedRow);
    } else {
      return evaluateConditionOnNamespacedRow(
        namespacedRow,
        item,
        mainTableAlias,
        joinedTableAlias
      );
    }
  });
}
function evaluateConditionOnNamespacedRow(namespacedRow, condition, mainTableAlias, joinedTableAlias) {
  if (condition.length === 3 && !Array.isArray(condition[0])) {
    const [left, comparator, right] = condition;
    return evaluateSimpleConditionOnNamespacedRow(
      namespacedRow,
      left,
      comparator,
      right,
      mainTableAlias,
      joinedTableAlias
    );
  }
  if (condition.length > 3 && !Array.isArray(condition[0]) && typeof condition[1] === `string` && ![`and`, `or`].includes(condition[1])) {
    let result = evaluateSimpleConditionOnNamespacedRow(
      namespacedRow,
      condition[0],
      condition[1],
      condition[2],
      mainTableAlias,
      joinedTableAlias
    );
    for (let i = 3; i < condition.length; i += 4) {
      const logicalOp = condition[i];
      if (i + 3 <= condition.length) {
        const nextResult = evaluateSimpleConditionOnNamespacedRow(
          namespacedRow,
          condition[i + 1],
          condition[i + 2],
          condition[i + 3],
          mainTableAlias,
          joinedTableAlias
        );
        if (logicalOp === `and`) {
          result = result && nextResult;
        } else {
          result = result || nextResult;
        }
      }
    }
    return result;
  }
  if (condition.length > 0 && Array.isArray(condition[0])) {
    let result = evaluateConditionOnNamespacedRow(
      namespacedRow,
      condition[0],
      mainTableAlias,
      joinedTableAlias
    );
    for (let i = 1; i < condition.length; i += 2) {
      if (i + 1 >= condition.length) break;
      const operator = condition[i];
      const nextCondition = condition[i + 1];
      if (operator === `and`) {
        result = result && evaluateConditionOnNamespacedRow(
          namespacedRow,
          nextCondition,
          mainTableAlias,
          joinedTableAlias
        );
      } else {
        result = result || evaluateConditionOnNamespacedRow(
          namespacedRow,
          nextCondition,
          mainTableAlias,
          joinedTableAlias
        );
      }
    }
    return result;
  }
  return true;
}
function evaluateSimpleConditionOnNamespacedRow(namespacedRow, left, comparator, right, mainTableAlias, joinedTableAlias) {
  const leftValue = evaluateOperandOnNamespacedRow(
    namespacedRow,
    left,
    mainTableAlias,
    joinedTableAlias
  );
  const rightValue = evaluateOperandOnNamespacedRow(
    namespacedRow,
    right,
    mainTableAlias,
    joinedTableAlias
  );
  switch (comparator) {
    case `=`:
      return leftValue === rightValue;
    case `!=`:
      return leftValue !== rightValue;
    case `<`:
      return compareValues(leftValue, rightValue, `<`);
    case `<=`:
      return compareValues(leftValue, rightValue, `<=`);
    case `>`:
      return compareValues(leftValue, rightValue, `>`);
    case `>=`:
      return compareValues(leftValue, rightValue, `>=`);
    case `like`:
    case `not like`:
      if (typeof leftValue === `string` && typeof rightValue === `string`) {
        const pattern = convertLikeToRegex(rightValue);
        const matches = new RegExp(`^${pattern}$`, `i`).test(leftValue);
        return comparator === `like` ? matches : !matches;
      }
      return comparator === `like` ? false : true;
    case `in`:
      if (!Array.isArray(rightValue)) {
        return false;
      }
      if (rightValue.length === 0) {
        return false;
      }
      if (Array.isArray(leftValue)) {
        return leftValue.some((item) => isValueInArray(item, rightValue));
      }
      return isValueInArray(leftValue, rightValue);
    case `not in`:
      if (!Array.isArray(rightValue)) {
        return true;
      }
      if (rightValue.length === 0) {
        return true;
      }
      if (Array.isArray(leftValue)) {
        return !leftValue.some((item) => isValueInArray(item, rightValue));
      }
      return !isValueInArray(leftValue, rightValue);
    case `is`:
      return leftValue === rightValue;
    case `is not`:
      if (rightValue === null) {
        return leftValue !== null && leftValue !== void 0;
      }
      return leftValue !== rightValue;
    default:
      return false;
  }
}
export {
  evaluateConditionOnNamespacedRow,
  evaluateSimpleConditionOnNamespacedRow,
  evaluateWhereOnNamespacedRow
};
//# sourceMappingURL=evaluators.js.map
